"""
Alice-Bob Integrator Environment (ABIE), created by Alice and Bob in the Moving Planets Around (MPA) project.

Features:
    1. The integrator takes input either from the commandline, or from the config file
    2. The integrator periodically stores data to HDF5
    3. The integrator supports restarting simulations


The MPA team, 2017-2018
"""
import argparse
import toml
from abie.integrator import Integrator
import numpy as np
import sys
from abie.data_io import DataIO


class ABIE(object):

    def __init__(self):
        # =================== CONSTANTS ==================
        # by default, using the square of the Gaussian gravitational constant
        self.__CONST_G = 0.000295912208232213  # units: (AU^3/day^2)
        self.__CONST_C = 0.0  # speed of light; PN terms will be calculated if CONST_C > 0

        # # =================== VARIABLES ==================
        self.__t = 0.0
        self.__t_start = 0.0
        self.__t_end = 0.0
        self.__h = 0.0
        self.__store_dt = 100  # output is triggered per 100 time units
        self.__buffer_len = 1024  # size of dataset in a hdf5 group
        self.__max_close_encounter_events = 1
        self.__max_collision_events = 1
        self.__close_encounter_distance = 0.0
        # self.acceleration_method = 'numpy'

        # load integrator modules
        self.__integrators = None  # a collection of integrators loaded from modules
        self.__integrator = None  # the actual, active integrator instance
        self.output_file = 'data.hdf5'
        self.__close_encounter_output_file = 'close_encounters.txt'
        self.__collision_output_file = 'collisions.txt'

    @property
    def max_close_encounter_events(self):
        if self.__integrator is not None:
            self.__max_close_encounter_events = self.__integrator.max_close_encounter_events
            return self.__max_close_encounter_events
        else:
            return self.__max_close_encounter_events

    @max_close_encounter_events.setter
    def max_close_encounter_events(self, value):
        self.__max_close_encounter_events = value
        if self.__integrator is not None:
            self.__integrator.max_close_encounter_events = value

    @property
    def max_collision_events(self):
        if self.__integrator is not None:
            self.__max_collision_events = self.__integrator.collision_events
            return self.__max_collision_events
        else:
            return self.__max_collision_events

    @max_collision_events.setter
    def max_collision_events(self, value):
        self.__max_collision_events = value
        if self.__integrator is not None:
            self.__integrator.max_collision_events = value

    @property
    def close_encounter_distance(self):
        if self.__integrator is not None:
            self.__close_encounter_distance = self.__integrator.max_close_encounter_events
            return self.__close_encounter_distance
        else:
            return self.__close_encounter_distance

    @close_encounter_distance.setter
    def close_encounter_distance(self, value):
        self.__close_encounter_distance = value
        if self.__integrator is not None:
            self.__integrator.close_encounter_distance = value

    @property
    def close_encounter_output_file(self):
        if self.__integrator is not None:
            self.__close_encounter_output_file = self.__integrator.close_encounter_output_file
            return self.__close_encounter_output_file
        else:
            return self.__close_encounter_output_file

    @close_encounter_output_file.setter
    def close_encounter_output_file(self, value):
        self.__close_encounter_output_file = value
        if self.__integrator is not None:
            self.__integrator.close_encounter_output_file = value

    @property
    def collision_output_file(self):
        if self.__integrator is not None:
            self.__collision_output_file = self.__integrator.collision_output_file
            return self.__collision_output_file
        else:
            return self.__collision_output_file

    @collision_output_file.setter
    def collision_output_file(self, value):
        self.__collision_output_file = value
        if self.__integrator is not None:
            self.__integrator.collision_output_file = value

    @property
    def integrator(self):
        if self.__integrator is None:
            raise RuntimeError('Integrator not set!')
        return self.__integrator

    @property
    def particles(self):
        if self.__integrator is None:
            raise RuntimeError('Particle sets undefined because the integrator is not set!')
        return self.__integrator.particles

    @property
    def t(self):
        if self.__integrator is not None:
            self.__t = self.integrator.t
            return self.__t
        else:
            return self.__t

    @property
    def CONST_G(self):
        if self.__integrator is not None:
            self.__CONST_G = self.__integrator.CONST_G
            return self.__CONST_G
        else:
            return self.__CONST_G

    @CONST_G.setter
    def CONST_G(self, value):
        self.__CONST_G = value
        if self.__integrator is not None:
            self.__integrator.CONST_G = value

    @property
    def CONST_C(self):
        if self.__integrator is not None:
            self.__CONST_C = self.__integrator.CONST_C
            return self.__CONST_C
        else:
            return self.__CONST_C

    @CONST_C.setter
    def CONST_C(self, value):
        self.__CONST_C = value
        if self.__integrator is not None:
            self.__integrator.CONST_C = value

    @property
    def t_end(self):
        if self.__integrator is not None:
            self.__t_end = self.integrator.t_end
            return self.__t_end
        else:
            return self.__t_end

    @t_end.setter
    def t_end(self, tf):
        self.__t_end = tf
        if self.__integrator is not None:
            self.integrator.t_end = tf

    @property
    def h(self):
        if self.__integrator is not None:
            self.__h = self.integrator.h
            return self.__h
        else:
            return self.__h

    @h.setter
    def h(self, value):
        self.__h = value
        if self.__integrator is not None:
            self.__integrator.h = value

    @property
    def store_dt(self):
        if self.__integrator is not None:
            self.__store_dt = self.__integrator.__store_dt
            return self.__store_dt
        else:
            return self.__store_dt

    @store_dt.setter
    def store_dt(self, value):
        self.__store_dt = value
        if self.__integrator is not None:
            self.__integrator.store_dt = self.__store_dt

    @property
    def buffer_len(self):
        if self.__integrator is not None:
            self.__buffer_len = self.__integrator.buffer_len
            return self.__buffer_len
        else:
            return self.__buffer_len

    @buffer_len.setter
    def buffer_len(self, value):
        self.__buffer_len = value
        if self.__integrator is not None:
            self.__integrator.buffer_len = self.__buffer_len

    @property
    def acceleration_method(self):
        return self.integrator.acceleration_method

    @acceleration_method.setter
    def acceleration_method(self, value):
        self.__integrator.acceleration_method = value

    @integrator.setter
    def integrator(self, name_of_integrator):
        if self.__integrators is None:
            self.__integrators = Integrator.load_integrators()
        print(('Setting the integrator to %s' % name_of_integrator))
        # populate the parameters to the integrator
        if name_of_integrator in self.__integrators:
            self.__integrator = getattr(self.__integrators[name_of_integrator], name_of_integrator)()
            self.__integrator.CONST_G = self.CONST_G
            self.__integrator.t_end = self.__t_end
            self.__integrator.h = self.__h
            self.__integrator.t_start = self.__t_start
            self.__integrator.output_file = self.output_file
            self.__integrator.collision_output_file = self.collision_output_file
            self.__integrator.close_encounter_output_file = self.close_encounter_output_file
            self.__integrator.store_dt = self.__store_dt
            self.__integrator.buffer_len = self.__buffer_len

    def initialize(self, config=None):
        # Initialize the integrator
        self.__integrators = Integrator.load_integrators()
        if self.__integrator is None:
            print('Use GaussRadau15 as the default integrator...')
            self.integrator = 'GaussRadau15'
            self.integrator.initialize()
            self.integrator.acceleration_method = 'ctypes'
        else:
            self.__integrator.CONST_G = self.CONST_G
            self.__integrator.t_end = self.__t_end
            self.__integrator.h = self.__h
            self.__integrator.t_start = self.__t_start
            self.__integrator.output_file = self.output_file
            self.__integrator.store_dt = self.__store_dt
            self.__integrator.buffer_len = self.__buffer_len


        if config is not None:
            # Gravitational parameter
            self.integrator.CONST_G = np.array(config['physical_params']['G'])

            # Integration parameters
            self.integrator = config['integration']['integrator']
            self.integrator.initialize()
            self.integrator.h = float(config['integration']['h'])
            if 'acc_method' in config['integration']:
                self.integrator.acceleration_method = config['integration']['acc_method']
            else:
                self.integrator.acceleration_method = 'ctypes'

            # Load sequence of object names
            if 'names' in config:
                names = config['names']
            else:
                names = None

            # Initial and final times
            if self.integrator.t_start == 0:
                self.integrator.t_start = float(config['integration']['t0'])
            if self.integrator.t_end == 0:
                self.integrator.t_end = float(config['integration']['tf'])
            self.integrator.active_integrator = config['integration']['integrator']
            DataIO.ic_populate(config['initial_conds'], self, names=names)

    def stop(self):
        """Stop the integrator and clean up the memory"""
        self.integrator.stop()

    def set_additional_forces(self, ext_acc):
        if ext_acc.ndim == 1 and ext_acc.shape[0] == 3 * self.integrator.particles.N:
            self.integrator.set_additional_forces(ext_acc)
        else:
            print('WARNING: Additional forces array needs to be 3 * N vector, where N is the number of particles.')

    def integrate(self, to_time=None):
        try:
            return self.integrator.integrate(to_time)
        except KeyboardInterrupt as e:
            print('Keyboard Interruption detected (Ctrl+C). Simulation stopped. Stopping the code...')
            self.stop()
            sys.exit(0)

    def calculate_orbital_elements(self, primary=None):
        return self.integrator.calculate_orbital_elements(primary)

    def calculate_energy(self):
        return self.integrator.calculate_energy()

    def add(self, pos=None, vel=None, x=None, y=None, z=None, vx=None, vy=None, vz=None, mass=0.0, name=None,
            radius=0.0, ptype=0, a=None, e=0.0, i=0.0, Omega=0.0, omega=0.0, f=0.0, primary=None):
        if x is not None and y is not None and z is not None:
            pos = np.empty(3, dtype=np.double)
            pos[0] = x
            pos[1] = y
            pos[2] = z
        if x is not None and y is not None and z is not None:
            pos = np.empty(3, dtype=np.double)
            pos[0] = x
            pos[1] = y
            pos[2] = z
        if vx is not None and vy is not None and vz is not None:
            vel = np.empty(3, dtype=np.double)
            vel[0] = vx
            vel[1] = vy
            vel[2] = vz
        return self.integrator.particles.add(pos=pos, vel=vel, mass=mass, name=name, radius=radius,
                                             ptype=ptype, a=a, e=e, i=i, Omega=Omega,
                                             omega=omega, f=f, primary=primary)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config', help='config file', default=None)
    parser.add_argument('-o', '--output_file', dest='output_file', help='output data file', default='data.hdf5')
    parser.add_argument('-r', '--rebound_file', help='Rebound simulation file', default=None)
    parser.add_argument('-t', '--t_end', type=float, dest='t_end', help='Termination time')
    parser.add_argument('-d', '--dt', type=float, dest='dt', help='Integration time step (optional for certain integrators)', default=None)
    parser.add_argument('-s', '--store_dt', type=float, dest='store_dt', help='output time step', default=100)
    parser.add_argument('-i', '--integrator', dest='integrator', help='Name of the integrator [GaussRadau15|WisdomHolman|RungeKutta|AdamsBashForth|LeapFrog|Euler]', default='GaussRadau15')

    args = parser.parse_args()
    abie = ABIE()
    abie.integrator = args.integrator
    if args.output_file is not None:
        abie.output_file = args.output_file
    if args.t_end is not None:
        abie.t_end = args.t_end
    if args.config is not None:
        abie.initialize(DataIO.parse_config_file(args.config))
    elif args.rebound_file is not None:
        # populate the initial conditions from rebound simulation files
        abie.initialize()
        DataIO.ic_populate_from_rebound(args.rebound_file, abie)
    if args.dt is not None:
        abie.h = args.dt
    abie.store_dt = args.store_dt
    abie.integrate()



if __name__ == "__main__":
    main()



from abie.integrator import Integrator
import numpy as np
from abie.events import *
from abie.ode import ODE
import sys

__integrator__ = 'WisdomHolman'


class WisdomHolman(Integrator):

    def __init__(self):
        super(self.__class__, self).__init__()

    def integrator_warmup(self):
        state_vec = np.concatenate((self.particles.positions, self.particles.velocities))
        helio = WisdomHolman.move_to_helio(state_vec, self.particles.N)
        pos = helio[0:3*self.particles.N].copy()
        vel = helio[3*self.particles.N:].copy()
        self.libabie.set_state(pos, vel, self.particles.masses, self.particles.radii, self.particles.N,
                               self.CONST_G, self.CONST_C)
        self.__energy_init = self.calculate_energy()


    def integrate_ctypes(self, to_time=None):
        ret = 0
        try:
            pos = np.empty(3 * self.particles.N)
            vel = np.empty(3 * self.particles.N)
            self.libabie.integrator_wh(self.t, to_time, self.h)

            self.libabie.get_state(pos, vel, self.particles.masses, self.particles.radii)
            self.particles.positions = pos
            self.particles.velocities = vel
            self._t = self.libabie.get_model_time()
            self.store_state()
            # self._t = to_time
        except CollisionException as e:
            print(e)
            self.store_collisions(self.libabie.get_collision_data())
            ret = 1
        except CloseEncounterException as e:
            print(e)
            self.store_close_encounters(self.libabie.get_close_encounter_data())
            ret = 2
        finally:
            self.libabie.reset_collision_buffer()

        return ret

    def integrate_numpy(self, to_time=None):
        # Generate state vector:
        x0 = np.concatenate((self._particles.positions, self._particles.velocities))

        # Time step (d):
        # dt = WisdomHolman.initial_step_size(x0, self._particles.masses, self.particles.N, 1, factor)

        # Move to a heliocentric frame
        helio = WisdomHolman.move_to_helio(x0, self.particles.N)
        self.particles.positions = helio[0:3*self.particles.N]
        self.particles.velocities = helio[3*self.particles.N:]
        # helio = x0.copy()

        # Initialize: compute Jacobi coordinates and initial acceleration:
        jacobi = WisdomHolman.helio2jacobi(helio, self.particles.masses, self.particles.N)
        accel = WisdomHolman.compute_accel(helio, jacobi, self.particles.masses, self.particles.N, self.CONST_G)
        # Compute energy:
        # energy_init = self.calculate_energy()
        # energy_init = WisdomHolman.compute_energy(helio, self.particles.masses, self.particles.N, self.CONST_G)
        t_current = self.t
        while t_current < to_time:
            helio, accel = WisdomHolman.wh_advance_step(helio, t_current, self.h, self.particles.masses, self.particles.N, accel, self.CONST_G)
            t_current += self.h
        self._t = to_time
        self.particles.positions = helio[0:3*self.particles.N]
        self.particles.velocities = helio[3*self.particles.N:]
        self.store_state()

        # while self.t < to_time:

            # Advance one step:
            # next_store_t = self.t + self.store_dt
                # bary = WisdomHolman.helio2bary(helio, self.particles.masses, self.particles.N)
            # helio = WisdomHolman.move_to_helio(helio, self.particles.N)
            # energy = self.calculate_energy()
            # print 'helio in np', helio
            # energy = WisdomHolman.compute_energy(helio, self.particles.masses, self.particles.N, self.CONST_G)
            # print('t = %f, E/E0 = %g' % (self.t, np.abs(energy-energy_init)/energy_init))
        # self.buf.close()
        return 0

    def calculate_energy(self):
        helio = np.concatenate((self.particles.positions, self.particles.velocities))
        return WisdomHolman.compute_energy(helio, self.particles.masses, self.particles.N, self.CONST_G)

    def propagate_wh(self, helio, t, h, store_dt, masses, N, accel, G):
        pass

    @staticmethod
    def propagate_kepler(t0, tf, vr0, vv0, gm):
        """
        Propagate Keplerian states using f and g functions

        :param t0: initial time
        :param tf: final time
        :param vr0: initial position vector
        :param vv0: initial velocity vector
        :param gm: gravitational parameter, G * (M + m)
        :return: vrf: final position vector; vvf: final velocity vector
        """
        # Check for trivial propagation:
        if (t0 == tf):
            vrf = vr0
            vvf = vv0
            return

        # Compute time step:
        dt = tf - t0

        # Internal tolerance for solving Kepler's equation:
        tol = 1e-12

        # Energy tolerance: used to distinguish between elliptic, parabolic, and hyperbolic orbits,
        # ideally 0:
        tol_energy = 0.0

        # Compute the magnitude of the initial position and velocity vectors:
        r0 = np.linalg.norm(vr0)
        v0 = np.linalg.norm(vv0)

        # Precompute sqrt(gm):
        sqrtgm = np.sqrt(gm)

        # Initial value of the Keplerian energy:
        xi = v0 ** 2 * .5 - gm / r0

        # Semimajor axis:
        sma = -gm / (2 * xi)
        alpha = 1 / sma

        if (alpha > tol_energy):
            # Elliptic orbits:
            chi0 = sqrtgm * dt * alpha
        elif (alpha < tol_energy):
            # Hyperbolic orbits:
            chi0 = np.sign(dt) * (np.sqrt(-sma) * np.log(-2 * gm * alpha * dt / (np.dot(vr0, vv0)
                                                                                 + np.sqrt(-gm * sma) * (
                                                                                 1 - r0 * alpha))))
        else:
            # Parabolic orbits:
            vh = np.cross(vr0, vv0)
            p = np.linalg.norm(vh) ** 2 / gm
            s = .5 * np.arctan(1 / (3 * np.sqrt(gm / p ** 3) * dt))
            w = np.arctan(np.tan(s) ** (1.0 / 3.0))
            chi0 = np.sqrt(p) * 2 / np.tan(2 * w)

        # Solve Kepler's equation:
        for j in range(0, 500):

            # Compute universal variable:
            psi = chi0 ** 2 * alpha

            # Compute C2 and C3:
            c2, c3 = WisdomHolman.compute_c2c3(psi)

            # Propagate radial distance:
            r = chi0 ** 2 * c2 + np.dot(vr0, vv0) / sqrtgm * chi0 * (1 - psi * c3) \
                + r0 * (1 - psi * c2)

            # Auxiliary variable for f and g functions:
            chi = chi0 + (sqrtgm * dt - chi0 ** 3 * c3 - np.dot(vr0, vv0) / sqrtgm * chi0 ** 2 * c2 \
                          - r0 * chi0 * (1 - psi * c3)) / r

            # Convergence:
            if (abs(chi - chi0) < tol):
                break

            chi0 = chi

        if (abs(chi - chi0) > tol):
            print("WARNING: failed to solver Kepler's equation, error = %23.15e\n" % abs(chi - chi0))

        # Compute f and g functions, together with their derivatives:
        f = 1 - chi ** 2 / r0 * c2
        g = dt - chi ** 3 / sqrtgm * c3
        dg = 1 - chi ** 2 / r * c2
        df = sqrtgm / (r * r0) * chi * (psi * c3 - 1)


        # Propagate states:
        vr = f * vr0 + g * vv0
        vv = df * vr0 + dg * vv0

        return vr, vv

    @staticmethod
    def compute_c2c3(psi):
        """
        Propagate Keplerian states using f and g functions

        :param psi: universal variable
        :return: c2, c3: auxiliary C2 and C3 functions
        """

        if (psi > 1e-10):
            c2 = (1 - np.cos(np.sqrt(psi))) / psi
            c3 = (np.sqrt(psi) - np.sin(np.sqrt(psi))) / np.sqrt(psi ** 3)

        else:
            if (psi < -1e-6):
                c2 = (1 - np.cosh(np.sqrt(-psi))) / psi
                c3 = (np.sinh(np.sqrt(-psi)) - np.sqrt(-psi)) / np.sqrt(-psi ** 3)
            else:
                c2 = 0.5
                c3 = 1.0 / 6.0

        return c2, c3


    @staticmethod
    def wh_advance_step(x, t, dt, masses, nbodies, accel, G):
        """
        Advance one step using the Wisdom-Holman mapping. Implements a Kick-Drift-Kick strategy.

        :param x: current state (heliocentric coordinates)
        :param t: current time
        :param dt: time step
        :param masses: masses of the bodies
        :param nbodies: number of bodies
        :param accel: acceleration from H_interaction
        :return:
            - helio: heliocentric state at t + dt
            - accel: updated acceleration at dt
        """

        # Create shallow copy:
        helio = x.copy()

        # Kick:
        helio = WisdomHolman.wh_kick(helio, dt / 2, masses, nbodies, accel)

        # Convert from heliocentric to Jacobi for drifting:
        jacobi = WisdomHolman.helio2jacobi(helio, masses, nbodies)

        # Drift
        jacobi = WisdomHolman.wh_drift(jacobi, dt, masses, nbodies, G)

        # Convert from Jacobi to heliocentric for kicking:
        helio = WisdomHolman.jacobi2helio(jacobi, masses, nbodies)

        # Compute acceleration at t + dt:
        accel = WisdomHolman.compute_accel(helio, jacobi, masses, nbodies, G)

        # Kick:
        helio = WisdomHolman.wh_kick(helio, dt / 2, masses, nbodies, accel)

        # Return the heliocentric coordinates:
        return helio, accel

    @staticmethod
    def wh_kick(x, dt, masses, nbodies, accel):
        """
        Apply momentum kick following the Wisdom-Holman mapping strategy.
        :param x: current state (heliocentric coordinates)
        :param dt: time step (local)
        :param masses: masses of the bodies
        :param nbodies: number of bodies
        :param accel: acceleration from H_interaction
        :return:
            - kick: state at t + dt after the kick
        """

        # Create shallow copy:
        kick = x.copy()

        # Change the momenta:
        kick[(nbodies + 1) * 3:] += accel[3:] * dt

        return kick

    @staticmethod
    def wh_drift(x, dt, masses, nbodies, G):
        """
        Drift, i.e. Keplerian propagation.

        :param x: current state (heliocentric coordinates)
        :param dt: time step (local)
        :param masses: masses of the bodies
        :param nbodies: number of bodies
        :return:
            - drift: state at t + dt after drift
        """

        # Drifted state:
        drift = np.zeros(nbodies * 6)

        # Propagate each body assuming Keplerian motion:
        eta0 = masses[0]
        for ibod in range(1, nbodies):
            # Interior mass:
            eta = eta0 + masses[ibod]

            # Compute equivalent GM:
            gm = G * masses[0] * eta / eta0

            # Initial conditions:
            pos0 = x[ibod * 3: (ibod + 1) * 3]
            vel0 = x[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3]

            # Propagate:
            pos, vel = WisdomHolman.propagate_kepler(0.0, dt, pos0, vel0, gm)

            # Store states:
            drift[ibod * 3: (ibod + 1) * 3] = pos
            drift[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3] = vel

            eta0 = eta

        return drift

    @staticmethod
    def helio2jacobi(x, masses, nbodies):
        """
        Transform from heliocentric to Jacobi coordinates.

        :param x: state in heliocentric coordinates
        :param masses: masses of the bodies
        :param nbodies: number of bodies
        :return:
            - jacobi: state in Jacobi coordinates
        """

        # Create shallow copy:
        jacobi = x.copy()

        # Compute etas (interior masses):
        eta = np.zeros(nbodies)
        eta[0] = masses[0]
        for ibod in range(1, nbodies):
            eta[ibod] = masses[ibod] + eta[ibod - 1]

        # Assume central body at rest:
        jacobi[0: 3] = 0.0
        jacobi[nbodies * 3: (nbodies + 1) * 3] = 0.0

        # Jacobi coordinates of first body coincide with heliocentric, leave as they are.

        # Compute internal c.o.m. and momentum:
        auxR = masses[1] * x[3: 6]
        auxV = masses[1] * x[(nbodies + 1) * 3: (nbodies + 2) * 3]
        Ri = auxR / eta[1]
        Vi = auxV / eta[1]
        for ibod in range(2, nbodies):
            jacobi[ibod * 3: (ibod + 1) * 3] = x[ibod * 3: (ibod + 1) * 3] - Ri
            jacobi[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3] = \
                x[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3] - Vi

            # Compute the next internal c.o.m. and momentum of the sequence:
            if (ibod < nbodies - 1):
                auxR += masses[ibod] * x[ibod * 3: (ibod + 1) * 3]
                auxV += masses[ibod] * x[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3]
                Ri = auxR / eta[ibod]
                Vi = auxV / eta[ibod]

        return jacobi

    @staticmethod
    def jacobi2helio(x, masses, nbodies):
        """
        Transform from Jacobi to heliocentric coordinates.

        :param x: state in Jacobi coordinates
        :param masses: masses of the bodies
        :param nbodies: number of bodies
        :return:
            - helio: state in heliocentric coordinates
        """

        # Create shallow copy:
        helio = x.copy()

        # Compute etas (interior masses):
        eta = np.zeros(nbodies)
        eta[0] = masses[0]
        for ibod in range(1, nbodies):
            eta[ibod] = masses[ibod] + eta[ibod - 1]
        # Assume central body at rest:
        helio[0: 3] = 0.0
        helio[nbodies * 3: (nbodies + 1) * 3] = 0.0

        # Heliocentric coordinates of first body coincide with Jacobi, leave as they are.

        # Compute internal c.o.m. and momentum:
        Ri = masses[1] * x[3: 6] / eta[1]
        Vi = masses[1] * x[(nbodies + 1) * 3: (nbodies + 2) * 3] / eta[1]
        for ibod in range(2, nbodies):
            helio[ibod * 3: (ibod + 1) * 3] = x[ibod * 3: (ibod + 1) * 3] + Ri
            helio[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3] = \
                x[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3] + Vi

            # Compute the next internal c.o.m. and momentum of the sequence:
            if (ibod < nbodies - 1):
                Ri += masses[ibod] * x[ibod * 3: (ibod + 1) * 3] / eta[ibod]
                Vi += masses[ibod] * x[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3] / eta[ibod]

        return helio

    @staticmethod
    def compute_accel(helio, jac, masses, nbodies, G):
        """
        Compute acceleration on all bodies.

        :param helio: current state in heliocentric coordinates
        :param jac: current state in Jacobi coordinates
        :param masses: masses of the bodies
        :param nbodies: number of bodies
        :return:
            - accel: acceleration vector
        """

        # Allocate:
        accel = np.zeros(nbodies * 3)

        # Acceleration of first body is assumed zero:
        # TODO: remove this assumption

        inv_r3helio = np.zeros(nbodies)
        inv_r3jac = np.zeros(nbodies)
        inv_rhelio = inv_r3helio
        inv_rjac = inv_r3jac
        for ibod in range(2, nbodies):
            inv_rhelio[ibod] = 1.0 / np.linalg.norm(helio[ibod * 3: (ibod + 1) * 3])
            inv_r3helio[ibod] = inv_rhelio[ibod] ** 3
            inv_rjac[ibod] = 1.0 / np.linalg.norm(jac[ibod * 3: (ibod + 1) * 3])
            inv_r3jac[ibod] = inv_rjac[ibod] ** 3

        # Compute all indirect terms at once:
        accel_ind = np.zeros(3)
        for ibod in range(2, nbodies):
            accel_ind -= G * masses[ibod] * helio[ibod * 3: (ibod + 1) * 3] * inv_r3helio[ibod]
        accel_ind = np.concatenate((np.zeros(3), np.tile(accel_ind, nbodies - 1)))

        # Compute contribution from central body:
        accel_cent = accel * 0.0
        for ibod in range(2, nbodies):
            accel_cent[ibod * 3: (ibod + 1) * 3] = G * masses[0] \
                                                   * (jac[ibod * 3: (ibod + 1) * 3] * inv_r3jac[ibod] \
                                                      - helio[ibod * 3: (ibod + 1) * 3] * inv_r3helio[ibod])

        # Compute third part of the Hamiltonian:
        accel2 = accel * 0.0
        etai = masses[0]
        for ibod in range(2, nbodies):
            etai += masses[ibod - 1]
            accel2[ibod * 3: (ibod + 1) * 3] = accel2[(ibod - 1) * 3: ibod * 3] \
                                               + G * masses[ibod] * masses[0] * inv_r3jac[ibod] / etai * jac[ibod * 3: (ibod + 1) * 3]

        # Compute final part of the Hamiltonian:
        accel3 = accel * 0.0
        for ibod in range(1, nbodies - 1):
            for jbod in range(ibod + 1, nbodies):
                diff = helio[jbod * 3: (jbod + 1) * 3] - helio[ibod * 3: (ibod + 1) * 3]
                aux = 1.0 / np.linalg.norm(diff) ** 3
                accel3[jbod * 3: (jbod + 1) * 3] -= G * masses[ibod] * aux * diff
                accel3[ibod * 3: (ibod + 1) * 3] += G * masses[jbod] * aux * diff

        # Add all contributions:
        accel = accel_ind + accel_cent + accel2 + accel3
        return accel

    @staticmethod
    def helio2bary(x, masses, nbodies):
        """
        Transform from heliocentric to barycentric coordinates.

        :param x: current state (heliocentric coordinates)
        :param masses: masses of the bodies
        :param nbodies: number of bodies
        :return:
            - bary: barycentric coordinates
        """

        # Total mass of the system:
        mtotal = masses.sum()

        # Allocate barycentric coordinates:
        bary = np.zeros(nbodies * 6)

        for ibod in range(1, nbodies):
            bary[0: 3] += masses[ibod] * x[ibod * 3: (ibod + 1) * 3]
            bary[nbodies * 3: (nbodies + 1) * 3] += masses[ibod] \
                                                    * x[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3]

        bary = -bary / mtotal

        for ibod in range(1, nbodies):
            bary[ibod * 3: (ibod + 1) * 3] = x[ibod * 3: (ibod + 1) * 3] + bary[0: 3]
            bary[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3] = \
                x[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3] \
                + bary[nbodies * 3: (nbodies + 1) * 3]

        return bary

    @staticmethod
    def move_to_helio(x, nbodies):
        helio = x.copy()
        for ibod in range(1, nbodies):
            helio[ibod * 3: (ibod + 1) * 3] = helio[ibod * 3: (ibod + 1) * 3] - helio[0: 3]
            helio[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3] = \
                helio[(nbodies + ibod) * 3: (nbodies + ibod + 1) * 3] \
                - helio[nbodies * 3: (nbodies + 1) * 3]
        return helio

    @staticmethod
    def compute_energy(helio, masses, nbodies, G):
        # Convert to barycentric:
        x = WisdomHolman.helio2bary(helio, masses, nbodies)

        pos = x[0: nbodies * 3]
        vel = x[nbodies * 3:]

        # Compute energy
        energy = 0.0
        for i in range(0, nbodies):
            energy += 0.5 * masses[i] * np.linalg.norm(x[(nbodies + i) * 3:(nbodies + 1 + i) * 3]) ** 2
            for j in range(0, nbodies):
                if (i == j):
                    continue
                energy -= .5 * G * masses[i] * masses[j] / np.linalg.norm(x[i * 3: 3 + i * 3] - x[j * 3: 3 + j * 3])

        return energy

    @staticmethod
    def initial_step_size(x, masses, nbodies, ibody, factor, G):
        """
        Initialize the step size.

        :param x: current state (heliocentric coordinates)
        :param masses: masses of the bodies
        :param nbodies: number of bodies
        :param ibody: id of the body whose orbital period will be used to compute the step size.
        :param factor: the initial step size will be the period of ibody divided by factor
        :return:
            - dt: step size
        """

        # Gravitational parameter:
        gm = G * masses[0] + masses[ibody]

        # Compute the relative distance:
        r = np.linalg.norm(x[ibody * 3: (ibody + 1) * 3])

        # Relative velocity:
        v = np.linalg.norm(x[(nbodies + ibody) * 3: (nbodies + ibody + 1) * 3])

        # Semimajor axis:
        sma = -gm / (v ** 2 - 2 * gm / r)

        # Orbital period:
        period = 2 * np.pi * np.sqrt(sma ** 3 / gm)

        # Initial step size:
        dt = period / factor

        return dt


import numpy as np


class Particle(object):
    def __init__(self, ptype=0, mass=0.0, pos=np.zeros(3), vel=np.zeros(3), radius=0.0, name=None, primary=None):
        self.ptype = ptype  # particle type. 0: regular particle; 1: massless; 2: low-mass
        self.mass = mass  # mass
        self.radius = radius  # radius
        self.name = name  # user-assigned name, optional
        # self.hash = hash(self)  # unique key of the particle
        self.hash = np.random.randint(100000000, 999999999)
        self.primary = primary  # this defines the primary object that it is orbiting
        self.__pos = pos
        self.__vel = vel
        self.x = pos[0]  # x
        self.y = pos[1]  # y
        self.z = pos[2]  # z
        self.vx = vel[0]  # vx
        self.vy = vel[1]  # vy
        self.vz = vel[2]  # vz

    def __repr__(self):
        return "Particle(m={0:g}, x={1:g}, y={2:g}, z={3:g}, vx={4:g}, vy={5:g}, vz={6:g}, r={7:g}, name='{8:s}', hash={9:d})".format(
            self.mass,
            self.x, self.y, self.z,
            self.vx, self.vy, self.vz,
            self.radius, self.name, self.hash)

    @property
    def pos(self):
        return self.__pos

    @property
    def vel(self):
        return self.__vel

    @pos.setter
    def pos(self, pos_vec):
        if type(pos_vec).__module__ == np.__name__:
            if pos_vec.size == 3:
                self.x = pos_vec[0]
                self.y = pos_vec[1]
                self.z = pos_vec[2]
                self.__pos = pos_vec
            else:
                raise ValueError('Position vector must be len=3 vector.')
        else:
            raise TypeError('Position vector must be a numpy vector with len=3.')

    @vel.setter
    def vel(self, vel_vec):
        if type(vel_vec).__module__ == np.__name__:
            if vel_vec.size == 3:
                self.vx = vel_vec[0]
                self.vy = vel_vec[1]
                self.vz = vel_vec[2]
                self.__vel = vel_vec
            else:
                raise ValueError('Velocity vector must be len=3 vector.')
        else:
            raise TypeError('Velocity vector must be a numpy vector with len=3.')



import ctypes
import os
import numpy as np
import sysconfig
from abie.events import CollisionException, CloseEncounterException


class CLibABIE(object):

    def __init__(self):
        __current_dir__ = os.path.dirname(os.path.realpath(__file__))
        lib_path = os.path.join(__current_dir__, 'libabie.so')

        suffix = sysconfig.get_config_var('EXT_SUFFIX')
        if suffix is None:
            suffix = ".so"
        lib_path = os.path.join(__current_dir__, '..', 'libabie'+suffix)
        if not os.path.isfile(lib_path):
            # try to find the library from the parent directory
            lib_path = os.path.join(os.path.abspath(os.path.join(__current_dir__, os.pardir)), 'libabie.so')

        if not os.path.isfile(lib_path):
            print('Warning! Shared library libabie.so not exsit! Trying to compile.')
            os.system('make')
        self.lib = ctypes.cdll.LoadLibrary(lib_path)
        self.max_close_encounter_events = 1
        self.max_collision_events = 1

    def initialize_code(self, G, C, N_MAX, MAX_CE_EVENTS=1, MAX_COLLISION_EVENTS=1, close_encounter_distance=0):
        self.lib.initialize_code(ctypes.c_double(G),
                                 ctypes.c_double(C),
                                 ctypes.c_int(N_MAX),
                                 ctypes.c_int(MAX_CE_EVENTS),
                                 ctypes.c_int(MAX_COLLISION_EVENTS))
        self.lib.set_close_encounter_distance(ctypes.c_double(close_encounter_distance))
        self.max_close_encounter_events = MAX_CE_EVENTS
        self.max_collision_events = MAX_COLLISION_EVENTS

    def finalize_code(self):
        self.lib.finalize_code()

    def set_state(self, pos, vel, masses, radii, N, G, C):
        self.lib.set_state(ctypes.c_void_p(pos.ctypes.data),
                           ctypes.c_void_p(vel.ctypes.data),
                           ctypes.c_void_p(masses.ctypes.data),
                           ctypes.c_void_p(radii.ctypes.data),
                           ctypes.c_int(N),
                           ctypes.c_double(G),
                           ctypes.c_double(C))

    def get_state(self, pos, vel, masses, radii):
        self.lib.get_state(ctypes.c_void_p(pos.ctypes.data),
                           ctypes.c_void_p(vel.ctypes.data),
                           ctypes.c_void_p(masses.ctypes.data),
                           ctypes.c_void_p(radii.ctypes.data))

    def get_model_time(self):
        self.lib.get_model_time.restype = ctypes.c_double
        return self.lib.get_model_time()

    def get_close_encounter_data(self):
        # each row: time, object 1, object 2, distance
        buf = np.zeros(4 * self.max_close_encounter_events)
        self.lib.get_close_encounter_buffer(ctypes.c_void_p(buf.ctypes.data))
        return buf.reshape(self.max_close_encounter_events, 4)

    def get_collision_data(self):
        # each row: time, object 1, object 2, distance
        buf = np.zeros(4 * self.max_collision_events)
        self.lib.get_collision_buffer(ctypes.c_void_p(buf.ctypes.data))
        return buf.reshape(self.max_collision_events, 4)

    def reset_close_encounter_buffer(self):
        self.lib.reset_close_encounter_buffer()

    def reset_collision_buffer(self):
        self.lib.reset_collision_buffer()

    def set_close_encounter_distance(self, value):
        self.lib.set_close_encounter_distance(ctypes.c_double(value))

    def get_total_energy(self):
        self.lib.calculate_energy.restype = ctypes.c_double
        return self.lib.calculate_energy()

    def set_additional_forces(self, ext_acc):
        """
        :param ext_acc: A 3 * N vector
        :return:
        """
        self.lib.set_additional_forces(ctypes.c_int(len(ext_acc)/3),
                                      ctypes.c_void_p(ext_acc.ctypes.data))

    def integrator_runge_kutta(self, pos, vel, masses, N, G, t, t_end, dt):
        self.lib.integrator_runge_kutta(ctypes.c_void_p(pos.ctypes.data),
                                        ctypes.c_void_p(vel.ctypes.data),
                                        ctypes.c_void_p(masses.ctypes.data),
                                        ctypes.c_int(N),
                                        ctypes.c_double(G),
                                        ctypes.c_double(t),
                                        ctypes.c_double(t_end),
                                        ctypes.c_double(dt))

    def integrator_gauss_radau15(self, pos, vel, masses, N, G, t, t_end, dt):
        self.lib.integrator_gauss_radau15(ctypes.c_void_p(pos.ctypes.data),
                                          ctypes.c_void_p(vel.ctypes.data),
                                          ctypes.c_void_p(masses.ctypes.data),
                                          ctypes.c_int(N),
                                          ctypes.c_double(G),
                                          ctypes.c_double(t),
                                          ctypes.c_double(t_end),
                                          ctypes.c_double(dt))

    def integrator_wisdom_holman(self, pos, vel, masses, N, G, t, t_end, dt):
        self.lib.integrator_wisdom_holman(ctypes.c_void_p(pos.ctypes.data),
                                          ctypes.c_void_p(vel.ctypes.data),
                                          ctypes.c_void_p(masses.ctypes.data),
                                          ctypes.c_int(N),
                                          ctypes.c_double(G),
                                          ctypes.c_double(t),
                                          ctypes.c_double(t_end),
                                          ctypes.c_double(dt))

    def integrator_gr(self, t, t_end, dt):
        self.lib.integrator_gr.restype = ctypes.c_int
        ret = self.lib.integrator_gr(ctypes.c_double(t), ctypes.c_double(t_end), ctypes.c_double(dt))
        if ret == 1:
            col_buf = self.get_close_encounter_data()
            raise CloseEncounterException(col_buf[-1, 0], int(col_buf[-1, 1]), int(col_buf[-1, 2]), col_buf[-1, 3])
        elif ret == 2:
            col_buf = self.get_collision_data()
            raise CollisionException(col_buf[-1, 0], int(col_buf[-1, 1]), int(col_buf[-1, 2]), col_buf[-1, 3])

    def integrator_rk(self, t, t_end, dt):
        self.lib.integrator_rk(ctypes.c_double(t), ctypes.c_double(t_end), ctypes.c_double(dt))

    def integrator_wh(self, t, t_end, dt):
        self.lib.integrator_wh(ctypes.c_double(t), ctypes.c_double(t_end), ctypes.c_double(dt))

    def ode_n_body_first_order(self, x, N, G, masses):
        dxdt = np.zeros(x.size, dtype=np.double)
        self.lib.ode_n_body_first_order(ctypes.c_void_p(x.ctypes.data),
                                        ctypes.c_int(N),
                                        ctypes.c_double(G),
                                        ctypes.c_void_p(masses.ctypes.data),
                                        ctypes.c_void_p(dxdt.ctypes.data))
        return dxdt

    def ode_n_body_second_order(self, x, N, G, masses):
        acc = np.zeros(x.size, dtype=np.double)
        self.lib.ode_n_body_second_order(ctypes.c_void_p(x.ctypes.data),
                                        ctypes.c_int(N),
                                        ctypes.c_double(G),
                                        ctypes.c_void_p(masses.ctypes.data),
                                        ctypes.c_void_p(acc.ctypes.data))
        return acc



import numpy as np


class Tools(object):

    @staticmethod
    def compute_energy(sol_state, masses, const_g):
        n_rows, n_cols = sol_state.shape
        if n_cols % 6 == 0 and n_rows > 0:
            energy = np.zeros(n_rows)
            for t in range(n_rows):
                for i in range(0, n_cols/6):
                    energy[t] += 0.5 * masses[i] * np.linalg.norm(sol_state[t][n_cols/2+i*3:n_cols/2+i*3+3]) ** 2
                    for j in range(0, n_cols/6):
                        if i == j:
                            continue
                        energy[t] -= .5 * const_g * masses[i] * masses[j] / np.linalg.norm(
                            sol_state[t][i*3:i*3+3] - sol_state[t][j*3:j*3+3])
            return energy

    @staticmethod
    def from_orbital_elements_to_cartesian(
            mp,
            ms,
            semimajor_axis,
            eccentricity,
            true_anomaly,
            inclination,
            argument_of_periapsis,
            longitude_of_ascending_node,
            G=4.0 * np.pi ** 2
    ):
        """

        Function that returns position and velocities computed from the input orbital
        elements. Angles in radians, inclination between 0 and 180

        """

        cos_true_anomaly = np.cos(true_anomaly)
        sin_true_anomaly = np.sin(true_anomaly)

        cos_inclination = np.cos(inclination)
        sin_inclination = np.sin(inclination)

        cos_arg_per = np.cos(argument_of_periapsis)
        sin_arg_per = np.sin(argument_of_periapsis)

        cos_long_asc_nodes = np.cos(longitude_of_ascending_node)
        sin_long_asc_nodes = np.sin(longitude_of_ascending_node)

        ### e_vec is a unit vector directed towards periapsis ###
        e_vec_x = cos_long_asc_nodes * cos_arg_per - sin_long_asc_nodes * sin_arg_per * cos_inclination
        e_vec_y = sin_long_asc_nodes * cos_arg_per + cos_long_asc_nodes * sin_arg_per * cos_inclination
        e_vec_z = sin_arg_per * sin_inclination
        e_vec = np.array([e_vec_x, e_vec_y, e_vec_z])

        ### q is a unit vector perpendicular to e_vec and the orbital angular momentum vector ###
        q_vec_x = -cos_long_asc_nodes * sin_arg_per - sin_long_asc_nodes * cos_arg_per * cos_inclination
        q_vec_y = -sin_long_asc_nodes * sin_arg_per + cos_long_asc_nodes * cos_arg_per * cos_inclination
        q_vec_z = cos_arg_per * sin_inclination
        q_vec = np.array([q_vec_x, q_vec_y, q_vec_z])

        #    print 'alpha',alphax**2+alphay**2+alphaz**2 # For debugging; should be 1
        #    print 'beta',betax**2+betay**2+betaz**2 # For debugging; should be 1

        ### Relative position and velocity ###
        separation = semimajor_axis * (1.0 - eccentricity ** 2) / (
                     1.0 + eccentricity * cos_true_anomaly)  # Compute the relative separation
        position_vector = separation * cos_true_anomaly * e_vec + separation * sin_true_anomaly * q_vec
        velocity_tilde = np.sqrt(G * (mp + ms) / (semimajor_axis * (1.0 - eccentricity ** 2)))  # Common factor
        velocity_vector = -1.0 * velocity_tilde * sin_true_anomaly * e_vec + velocity_tilde * (
                          eccentricity + cos_true_anomaly) * q_vec

        return position_vector, velocity_vector

    @staticmethod
    def from_cartesian_to_orbital_elements(mp, ms, position, velocity, G=4.0 * np.pi ** 2):
        """

        Function that computes orbital elements from cartesian coordinates.
        Return values are: mass1, mass2, semimajor axis, eccentricity,
        true anomaly, inclination, longitude of the ascending nodes and the
        argument of pericenter. All angles are returned in radians.
        In case of a perfectly circular orbit the true anomaly
        and argument of pericenter cannot be determined; in this case, the
        return values are 0.0 for both angles.

        """

        total_mass = mp + ms

        ### specific energy ###
        v_sq = np.dot(velocity, velocity)
        r_sq = np.dot(position, position)
        r = np.sqrt(r_sq)

        specific_energy = (1.0 / 2.0) * v_sq - G * total_mass / r
        # if specific_energy >= 0.0:
        #     print 'Not a bound orbit!'

        semimajor_axis = -G * total_mass / (2.0 * specific_energy)

        ### specific angular momentum ###
        specific_angular_momentum = np.cross(position, velocity)
        specific_angular_momentum_norm = np.sqrt(np.dot(specific_angular_momentum, specific_angular_momentum))
        specific_angular_momentum_unit = specific_angular_momentum / specific_angular_momentum_norm

        maximum_specific_angular_momentum_norm = G * total_mass / (np.sqrt(-2.0 * specific_energy))
        ell = specific_angular_momentum_norm / maximum_specific_angular_momentum_norm  ### specific AM in units of maximum AM

        ### for e = 0 or e nearly 0, ell can be slightly larger than unity due to numerical reasons ###
        ell_epsilon = 1e-15

        completely_or_nearly_circular = False

        if ell > 1.0:
            if 1.0 < ell <= ell + ell_epsilon:  ### still unity within numerical precision
                print('orbit is completely or nearly circular; in this case the LRL vector cannot be used to reliably obtain the argument of pericenter and true anomaly; the output values of the latter will be set to zero; output e will be e = 0')
                ell = 1.0
                completely_or_nearly_circular = True
            else:  ### larger than unity within numerical precision
                raise Exception("angular momentum larger than maximum angular momentum for bound orbit")

        eccentricity = np.sqrt(1.0 - ell ** 2)

        ### Orbital inclination ###
        z_vector = np.array([0., 0., 1.])
        inclination = np.arccos(np.dot(z_vector, specific_angular_momentum_unit))

        ### Longitude of ascending nodes, with reference direction along x-axis ###
        ascending_node_vector = np.cross(z_vector, specific_angular_momentum)
        ascending_node_vector_norm = np.sqrt(np.dot(ascending_node_vector, ascending_node_vector))
        if ascending_node_vector_norm == 0:
            ascending_node_vector_unit = np.array([1., 0., 0.])
        else:
            ascending_node_vector_unit = ascending_node_vector / ascending_node_vector_norm

        long_asc_nodes = np.arctan2(ascending_node_vector_unit[1], ascending_node_vector_unit[0])

        ### Argument of periapsis and true anomaly, using eccentricity a.k.a. Laplace-Runge-Lenz (LRL) vector ###
        mu = G * total_mass
        position_unit = position / r
        e_vector = (1.0 / mu) * np.cross(velocity,
                                            specific_angular_momentum) - position_unit  ### Laplace-Runge-Lenz vector

        if completely_or_nearly_circular == True:  ### orbit is completely or nearly circular; in this case the LRL vector cannot be used to reliably obtain the argument of pericenter and true anomaly; the output values of the latter will be set to zero; output e will be e = 0
            arg_per = 0.0
            true_anomaly = 0.0
        else:
            e_vector_norm = np.sqrt(np.dot(e_vector, e_vector))
            e_vector_unit = e_vector / e_vector_norm

        e_vector_unit_cross_AM_unit = np.cross(e_vector_unit, specific_angular_momentum_unit)
        sin_arg_per = np.dot(ascending_node_vector_unit, e_vector_unit_cross_AM_unit)
        cos_arg_per = np.dot(e_vector_unit, ascending_node_vector_unit)
        arg_per = np.arctan2(sin_arg_per, cos_arg_per)

        sin_true_anomaly = np.dot(position_unit, -1.0 * e_vector_unit_cross_AM_unit)
        cos_true_anomaly = np.dot(position_unit, e_vector_unit)
        true_anomaly = np.arctan2(sin_true_anomaly, cos_true_anomaly)

        return semimajor_axis, eccentricity, true_anomaly, inclination, arg_per, long_asc_nodes

    @staticmethod
    def from_cartesian_to_aei(mp, ms, position, velocity, G=0.0172020989**2, z_vector=np.array([0., 0., 1.])):
        """

        Function that computes orbital elements from cartesian coordinates.
        Return values are: mass1, mass2, semimajor axis, eccentricity,
        true anomaly, inclination, longitude of the ascending nodes and the
        argument of pericenter. All angles are returned in radians.
        In case of a perfectly circular orbit the true anomaly
        and argument of pericenter cannot be determined; in this case, the
        return values are 0.0 for both angles.

        """

        total_mass = mp + ms
        mu = G * total_mass

        if velocity.ndim == 1 and position.ndim == 1:
            # Calculate the a-e-i of one body wrt another one
            v_sq = np.dot(velocity, velocity)
            r_sq = np.dot(position, position)
            r = np.sqrt(r_sq)
            if r == 0:
                return np.nan, np.nan, np.nan

            specific_energy = (1.0 / 2.0) * v_sq - G * total_mass / r
            # if specific_energy >= 0.0:
            #     print 'Not a bound orbit!'

            semimajor_axis = -G * total_mass / (2.0 * specific_energy)

            specific_angular_momentum = np.cross(position, velocity)
            specific_angular_momentum_norm = np.sqrt(np.dot(specific_angular_momentum, specific_angular_momentum))
            specific_angular_momentum_unit = specific_angular_momentum / specific_angular_momentum_norm
            ecc_vec = ((v_sq - mu / r) * position - np.dot(position, velocity) * velocity) / mu

            inclination = np.arccos(np.dot(z_vector, specific_angular_momentum_unit))
            eccentricity = np.linalg.norm(ecc_vec)

        elif velocity.ndim == 2 and position.ndim == 2:

            # calculate a list of bodies (pos and vel are n*3 arrays, representing n points)
            v_sq = np.sum(velocity * velocity, axis=0)
            r_sq = np.sum(position * position, axis=0)
            r = np.sqrt(r_sq)

            specific_energy = (1.0 / 2.0) * v_sq - G * total_mass / r

            semimajor_axis = -G * total_mass / (2.0 * specific_energy)

            ### specific angular momentum ###
            specific_angular_momentum = np.cross(position.transpose(), velocity.transpose()).transpose()
            specific_angular_momentum_norm = np.sqrt(
                np.sum(specific_angular_momentum * specific_angular_momentum, axis=0))
            specific_angular_momentum_unit = specific_angular_momentum / specific_angular_momentum_norm
            ecc_vec = ((v_sq - mu / r) * position - np.sum(position * velocity, axis=0) * velocity) / mu

            eccentricity = np.linalg.norm(ecc_vec, axis=0)

            ### Orbital inclination ###
            inclination = np.arccos(np.matmul(z_vector, specific_angular_momentum_unit))

        return semimajor_axis, eccentricity, inclination



class ParticleException(Exception):
    """
    Base class for the particle exception events.
    """
    def __init__(self, t, obj1, obj2, distance):
        self.t = t
        self.obj1 = obj1
        self.obj2 = obj2
        self.distance = distance


class CloseEncounterException(ParticleException):
    """
    Exception for the close encounter events between two particles.
    """
    def __str__(self):
        return 'Close encounter detected between particle #%d and particle #%d with a distance of %g at t = %g' % (self.obj1, self.obj2, self.distance, self.t)


class CollisionException(ParticleException):
    """
    Exception for the collision events between two particles.
    """

    def __str__(self):
        return 'Collision detected between particle #%d and particle #%d with a distance of %g at t = %g' % (self.obj1, self.obj2, self.distance, self.t)



from abie.integrator import Integrator
from abie.ode import ODE
import numpy as np

__integrator__ = 'Euler'


class Euler(Integrator):

    def __init__(self):
        super(Euler, self).__init__()
        self.__initialized = False

    def integrate(self, to_time=None):
        if self.__initialized is False:
            self.initialize()
            self.__initialized = True
        # Allocate dense output
        npts = int(np.floor((self.t_end - self.t_start) / self.h) + 1)

        # Initial state
        x = np.concatenate((self._particles.positions, self._particles.velocities))
        # Vector of times
        sol_time = np.linspace(self.t_start, self.t_start + self.h * (npts - 1), npts)
        energy_init = self.calculate_energy()
        # Launch integration
        count = 1
        for t in sol_time[count:]:
            dxdt = ODE.ode_n_body_first_order(x, self.CONST_G, self._particles.masses)
            # Advance step
            x = x + dxdt * self.h

            # Store solution
            self._particles.positions = x[0:self._particles.N * 3]
            self._particles.velocities = x[self._particles.N * 3:]
            count += 1
            self._t = t
            self.store_state()
            energy = self.calculate_energy()
            print(('t = %f, E/E0 = %g' % (self.t, np.abs(energy - energy_init) / energy_init)))
        self.buf.close()
        return 0



"""
Alice-Bob Integrator (ABI), created by Alice and Bob in the Moving Planets Around (MPA) project.

Features:
    1. The integrator takes input either from the commandline, or from the config file
    2. The integrator periodically stores data to HDF5
    3. The integrator supports restarting simulations


The MPA team, 2017-2018
"""

import glob
import os
import sys
import numpy as np
from abie.particles import Particles
from abie.clibabie import CLibABIE
from abie.data_io import DataIO

__mpa_dir__ = os.path.dirname(os.path.abspath(__file__))
__user_shell_dir__ = os.getcwd()


class Integrator(object):

    def __init__(self):
        """
        The constructor of an abstract integrator
        """
        # =================== CONSTANTS ==================
        # by default, using the square of the Gaussian gravitational constant
        self.CONST_G = 0.000295912208232213  # units: (AU^3/day^2)
        self.CONST_C = 0.0  # speed of light; PN terms will be calculated if CONST_C > 0

        # =================== VARIABLES ==================
        self._t = 0.0
        self.t_start = 0.0
        self.t_end = 0.0
        self.h = 0.01  # time step size
        self.store_dt = 100  # storage time step
        self._particles = None
        self.acceleration_method = 'ctypes'
        self.output_file = 'data.hdf5'
        self.collision_output_file = 'collisions.txt'
        self.close_encounter_output_file = 'close_encounters.txt'
        self.max_close_encounter_events = 1
        self.max_collision_events = 1
        self.close_encounter_distance = 0.0
        self.__energy_init = 0.0
        self.__energy = 0.0
        self.__buf = None
        self.buffer_len = 1024
        self.__initialized = False

        # =============== C Library =============
        self.libabie = CLibABIE()

    @property
    def t(self):
        return self._t

    @property
    def particles(self):
        if self._particles is None:
            self._particles = Particles(self.CONST_G)
        return self._particles

    @property
    def buf(self):
        if self.__buf is None:
            self.__buf = DataIO(buf_len=self.buffer_len,
                                output_file_name=self.output_file,
                                CONST_G=self.CONST_G)
        return self.__buf

    @staticmethod
    def load_integrators():
        """
        Load integrator modules
        :return: a dict of integrator class objects, mapping the name of the integrator to the class object
        """
        mod_dict = dict()
        module_candidates = glob.glob(os.path.join(__mpa_dir__, 'integrator_*.py'))
        sys.path.append(__mpa_dir__)  # append the python path
        if __mpa_dir__ != __user_shell_dir__:
            # load the integrator module (if any) also from the current user shell directory
            module_cwd = glob.glob(os.path.join(__user_shell_dir__, 'integrator_*.py'))
            for m_cwd in module_cwd:
                module_candidates.append(m_cwd)
            sys.path.append(__user_shell_dir__)  # append the python path
        for mod_name in module_candidates:
            mod_name = os.path.basename(mod_name)
            mod = __import__(mod_name.split('.')[0])
            if hasattr(mod, '__integrator__'):
                # it is a valid ABI module, register it as a module
                mod_dict[mod.__integrator__] = mod
        return mod_dict

    def initialize(self):
        if self.__buf is None:
            self.__buf = DataIO(buf_len=self.buffer_len,
                                output_file_name=self.output_file,
                                close_encounter_output_file_name=self.close_encounter_output_file,
                                collision_output_file_name=self.collision_output_file,
                                CONST_G=self.CONST_G)
        if self.particles.N > 0:
            # initialize the C library
            self.libabie.initialize_code(self.CONST_G, self.CONST_C, self.particles.N,
                                         MAX_CE_EVENTS=self.max_close_encounter_events,
                                         MAX_COLLISION_EVENTS=self.max_collision_events,
                                         close_encounter_distance=self.close_encounter_distance)
            self.buf.initialize_buffer(self.particles.N)

    def stop(self):
        if self.__buf is not None:
            self.__buf.close()

    def calculate_orbital_elements(self, primary=None):
        return self._particles.calculate_orbital_elements(primary)

    def calculate_energy(self):
        # return self._particles.energy
        if self.acceleration_method == 'ctypes':
            return self.libabie.get_total_energy()
        else:
            return self._particles.energy

    def set_additional_forces(self, ext_acc):
        self.libabie.set_additional_forces(ext_acc)

    def integrator_warmup(self):
        pos = self.particles.positions.copy()
        vel = self.particles.velocities.copy()
        self.libabie.set_state(pos, vel, self.particles.masses, self.particles.radii, self.particles.N,
                               self.CONST_G, self.CONST_C)

    def integrate(self, to_time=None):
        """
        Integrate the system to a given time.
        :param to_time: The termination time. If None, it will use the self.t_end value, and the code will be stopped
                        when reaching self.t_end (i.e. if this function is called without argument, it can only be called
                        once; but if it is called with a to_time argument specificed, then it can be called iteratively.
        :return:
        """
        if to_time is not None:
            self.t_end = to_time

        if self.__initialized is False:
            self.initialize()
            self.integrator_warmup()
            self.__initialized = True
        if self.t >= self.t_end:
            return

        # Note: this dt is not the integrator time step h
        dt = min(self.store_dt, self.t_end-self.t)

        ret = 0
        # launch the integration
        while self.t < self.t_end:
            if self.__energy_init == 0:
                self.__energy_init = self.calculate_energy()
            next_t = self.t + dt - ((self.t + dt) % dt)
            if self.acceleration_method == 'numpy':
                ret = self.integrate_numpy(next_t)
            elif self.acceleration_method == 'ctypes':
                ret = self.integrate_ctypes(next_t)
            # the self.t is updated by the subclass
            # energy check
            self.__energy = self.calculate_energy()
            print(('t = %f, N = %d, dE/E0 = %g' % (self.t, self.particles.N, np.abs(self.__energy - self.__energy_init) / self.__energy_init)))
            if os.path.isfile('STOP'):
                break

        if to_time is None:
            # triggering the termination of the code, save the buffer to the file and close it
            self.stop()
            # if ret > 0:
            #     break
        # if self.t == self.t_end:
        #     self.__energy = self.calculate_energy()
        #     print('t = %f, E/E0 = %g' % (self.t, np.abs(self.__energy - self.__energy_init) / self.__energy_init))
        return ret

    def integrate_numpy(self, to_time):
        """
        Integrate the system to a given time using python/numpy.
        This method must be implemented in the subclasses.
        :param to_time: The termination time. If None, it will use the self.t_end value
        :return:
        """
        raise NotImplementedError('integrate_numpy() method not implemented!')

    def integrate_ctypes(self, to_time):
        """
        Integrate the system to a given time using the ctypes (libabie.so)
        This method must be implemented in the subclasses.
        :param to_time: The termination time. If None, it will use the self.t_end value
        :return:
        """
        raise NotImplementedError('integrate_ctypes() method not implemented!')

    def store_state(self):
        if self.buf is None:
            self.initialize()
        self.buf.initialize_buffer(self.particles.N)
        elem = self.particles.calculate_aei()
        self.buf.store_state(self.t, self.particles.positions, self.particles.velocities, self.particles.masses,
                             radii=self.particles.radii, names=self.particles.hashes, ptypes=self.particles.ptypes,
                             a=elem[:, 0], e=elem[:, 1], i=elem[:, 2])

    def store_collisions(self, collision_buffer):
        self.buf.store_collisions(collision_buffer)

    def store_close_encounters(self, ce_buffer):
        self.buf.store_close_encounters(ce_buffer)

    def handle_collisions(self, collision_buffer, actions=None):
        if actions is None:
            actions = ['merge', 'store']
        if 'store' in actions:
            self.store_state()
            self.store_collisions(collision_buffer)
        if 'merge' in actions:
            collision_buffer = collision_buffer.reshape(len(collision_buffer), 4)
            for coll_pair in range(collision_buffer.shape[0]):
                pid1 = int(collision_buffer[coll_pair, 1])
                pid2 = int(collision_buffer[coll_pair, 2])
                self.particles.merge_particles_inelastically(pid1, pid2)
                self.libabie.reset_collision_buffer()
                self.integrator_warmup()
                self.buf.flush()
                self.buf.reset_buffer()
                self.buf.initialize_buffer(self.particles.N)
        if 'halt' in actions:
            print('Simulation terminated due to a collision event.')
            sys.exit(0)

    def handle_close_encounters(self, ce_buffer, actions=None):
        if actions is None:
            actions = ['merge', 'store']
        if 'store' in actions:
            self.store_state()
            self.store_close_encounters(ce_buffer)
        if 'halt' in actions:
            print('Simulation terminated due to a collision event.')
            sys.exit(0)



import os
import sys
import h5py
import glob
import shutil
import numpy as np
from optparse import OptionParser


parser = OptionParser()
parser.add_option('-f', '--file', action='store', type='string', dest='file',
                  help='The hdf5 file to be serialized (e.g., p_sys_0.hdf5.unfinished)')
parser.add_option('--dataset', action='store', type='string', dest='dataset', default=None,
                  help='The dataset of interest to be serialized (e.g., mass). If not specified, serialize all datasets.')
# parser.add_option('-o', '--output_file', action='store', type='string', dest='output_file',
#                   help='The name of the output file (optional)')
(options, args) = parser.parse_args()
if options.file is None:
    print('Usage: python snapshot_serialization -f <file_to_be_serialized.hdf5>')
    sys.exit(0)

h5fns = glob.glob(options.file)
print(h5fns)
if len(h5fns) > 0:
    for h5fn_id, h5fn in enumerate(h5fns):
        output_file_name = os.path.splitext(os.path.basename(h5fn))[0]
        print(('Processing %s' % h5fn))
        # temporarily copying the .unfinished file to /tmp
        new_file_path = os.path.join('/tmp', os.path.basename(h5fn))
        shutil.copyfile(h5fn, new_file_path)
        with h5py.File(new_file_path, 'r') as h5f, h5py.File(output_file_name, 'w') as h5f_out:
            step_id_list = []
            step_len_vec = np.zeros(len(h5f), dtype=np.int)
            if len(h5f['/Step#0/hash']) > 0:
                vec_hash = h5f['/Step#0/hash'].value[0]  # the hash array of the first step (which contains all particles)
            else:
                vec_hash = h5f['/Step#1/hash'].value[0]  # the hash array of the first step (which contains all particles)
            print('vec_hash', vec_hash.shape)
            dset_dict = dict()
            print('Mapping internal data structure...')
            for dset_name in h5f:
                if 'Step#' in dset_name:
                    step_id = int(dset_name.split('#')[1])
                    step_id_list.append(step_id)
                    step_len_vec[step_id] = (h5f['/%s/%s' % (dset_name, 'x')].value.shape[0])
                else:
                    # it is a serialized dataset, just copy it to the serialized file
                    h5f_out.create_dataset(dset_name, data=h5f[dset_name].value)

            if len(step_id_list) > 0:  # if more than zero steps
                cursor = 0
                for step_id in sorted(step_id_list):
                    step_name = 'Step#%d' % step_id
                    print(step_name)
                    h5g = h5f[step_name]
                    if len(h5g['hash'].value)  == 0:
                        continue
                    indices = np.where(np.in1d(vec_hash, h5g['hash'].value))[0] # the indices to be updated

                    if options.dataset is not None:
                        dset_list = [options.dataset]
                    else:
                        dset_list = h5g.keys()

                    for dset_name in dset_list:
                        print(dset_name)
                        tmp_data = h5g[dset_name].value
                        lower_idx = cursor
                        higher_idx = cursor + tmp_data.shape[0]
                        print('cursors', lower_idx, higher_idx, cursor, step_len_vec[step_id])
                        if dset_name not in dset_dict.keys():
                            # allocate memory
                            if tmp_data.ndim == 2:
                                dset_dict[dset_name] = np.zeros((np.sum(step_len_vec), tmp_data.shape[1])) * np.nan
                            elif h5g[dset_name].value.ndim == 1:
                                dset_dict[dset_name] = np.zeros(np.sum(step_len_vec)) * np.nan
                        # append the data to the allocated big array
                        # dset_dict[dset_name][step_id*tmp_data.shape[0]:(step_id+1)*tmp_data.shape[0]][indices] = tmp_data
                        if tmp_data.ndim == 2:
                            print(dset_name, tmp_data.shape, dset_dict[dset_name][lower_idx:higher_idx].shape)
                            print(indices, vec_hash.shape)
                            dset_dict[dset_name][lower_idx:higher_idx, :][:, indices] = tmp_data
                        else:
                            # print dset_name, tmp_data.shape, dset_dict[dset_name][lower_idx:higher_idx].shape
                            dset_dict[dset_name][lower_idx:higher_idx] = tmp_data
                    cursor += step_len_vec[step_id]
                        # dset_dict[dset_name][step_id][indices] = tmp_data
                for dset_name in dset_dict.keys():
                    if dset_name == 'hash':
                        h5f_out.create_dataset(dset_name, data=dset_dict[dset_name], dtype=np.int64)
                    else:
                        h5f_out.create_dataset(dset_name, data=dset_dict[dset_name])
        os.remove(new_file_path)



import numpy as np
import ctypes
import os
# from clibabie import CLibABIE

# if not os.path.isfile('libabie.so'):
#     print('Warning! Shared library libabie.so not exsit! Trying to compile.')
#     os.system('make')
# lib = ctypes.cdll.LoadLibrary("./libabie.so")
# fun_ode_first_order = lib.ode_n_body_first_order
# fun_ode_second_order = lib.ode_n_body_second_order
# fun.restype = None
# fun.argtypes = [ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
#                 ctypes.c_size_t,
#                 ctypes.c_double,
#                 ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
#                 ndpointer(ctypes.c_double, flags="C_CONTIGUOUS")]


# libabie = CLibABIE()


class ODE(object):

    # @staticmethod
    # def ode_n_body_second_order(x, const_g, masses):
    #     dxdt = libabie.ode_n_body_second_order(x, x.size/3, const_g, masses)
    #     return dxdt

    @staticmethod
    def ode_n_body_first_order(x, const_g, masses):
        """
        TWO-BODY EQUATIONS
        :param x: the flattened state vector. The first half of x is the positions; the second half is velocities.
        :param const_g:
        :param masses:
        :return: the first derivative of x; the first half is velocities; the second half is accelerations.
        """
        # Allocate
        dxdt = x * 0.0

        # Compute additional perturbations
        # perturbation = compute_perturbation(x, t, G, masses, nbodies)

        # Differential equations:
        # - Position
        nbodies = x.size / 6
        dxdt[0:nbodies*3] = x[nbodies*3:]  # velocities
        for j in range(0, nbodies):
            # dxdt[j * 3:3 + j * 3] = x[nbodies * 3 + j * 3:nbodies * 3 + 3 + j * 3]
            Rj = x[j * 3:3 + j * 3]
            aj = Rj * 0
            for k in range(0, nbodies):
                if j == k:
                    continue
                Rk = x[k * 3:k * 3 + 3]
                rel_sep = Rj - Rk
                aj += (-const_g * masses[k] * rel_sep / np.linalg.norm(rel_sep) ** 3)
            dxdt[nbodies * 3 + j * 3:nbodies * 3 + 3 + j * 3] = aj  # accelerations

        # Add extra accelerations
        # dxdt[nbodies * 3:] += perturbation
        return dxdt

    @staticmethod
    def ode_n_body_second_order(x, const_g, masses):

        nbodies = x.size / 3  # WARNING: this x contains only positions!!

        # Allocate
        acc = x * 0.0

        # Compute additional perturbations
        # perturbation = compute_perturbation(x, t, G, masses, nbodies)

        # Differential equations:
        # - Position
        for j in range(0, nbodies):
            Rj = x[j * 3:3 + j * 3]
            aj = Rj * 0
            for k in range(0, nbodies):
                if j == k:
                    continue
                Rk = x[k * 3:3 + k * 3]
                rel_sep = Rj - Rk
                aj += -const_g * masses[k] * rel_sep / np.linalg.norm(rel_sep) ** 3
            acc[j * 3:3 + j * 3] = aj

        # Add extra accelerations
        # acc += perturbation

        return acc



import numpy as np
from abie.particle import Particle
from abie.tools import Tools
from six import string_types


class Particles(object):
    """
    A particle container.
    """

    def __init__(self, const_g):
        self.__particles = []
        self.__positions = np.array([])
        self.__velocities = np.array([])
        self.__masses = np.array([])
        self.__names = dict()
        self.__N = 0
        self.CONST_G = const_g
        self.primary = '#COM#'  # '#COM#', '#M_MAX#', '#M_MIN#', or name/ID

    def __repr__(self):
        str_concat = 'Total number of particles: %d\n' % len(self.__particles)
        for pid, p in enumerate(self.__particles):
            str_concat += (p.__repr__() + '\n')
        return str_concat

    @property
    def N(self):
        return self.__N

    @property
    def particles(self):
        return self

    @property
    def positions(self):
        return self.__positions

    @property
    def names(self):
        name_list = []
        for p in self.__particles:
            name_list.append(p.name)
        return name_list

    @property
    def hashes(self):
        hash_list = []
        for p in self.__particles:
            hash_list.append(p.hash)
        return hash_list

    @property
    def radii(self):
        radii_list = np.empty(self.__N)
        for p_id, p in enumerate(self.__particles):
            radii_list[p_id] = p.radius
        return radii_list

    @property
    def ptypes(self):
        ptype_list = []
        for p in self.__particles:
            ptype_list.append(p.ptype)
        return ptype_list

    @positions.setter
    def positions(self, pos_vec):
        """
        Set the positions of all particles with a flattened position vector.
        :param pos_vec: The position vector
        :return:
        """
        if type(pos_vec).__module__ == np.__name__:
            if pos_vec.size == 3 * self.__N:
                for p_id in range(self.__N):
                    self.__particles[p_id].pos = pos_vec[3*p_id:3*p_id+3]
                self.__positions = pos_vec
            else:
                raise ValueError('Position vector must be len=3 vector.')
        else:
            raise TypeError('Position vector must be a numpy vector with len=3*N.')

    @property
    def velocities(self):
        return self.__velocities

    @velocities.setter
    def velocities(self, vel_vec):
        """
        Set the velocities of all particles with a flattened velocity vector.
        :param vel_vec: The velocity vector
        :return:
        """
        if type(vel_vec).__module__ == np.__name__:
            if vel_vec.size == 3 * self.__N:
                for p_id in range(self.__N):
                    self.__particles[p_id].vel = vel_vec[3*p_id:3*p_id+3]
                self.__velocities = vel_vec
            else:
                raise ValueError('Velocity vector must be len=3*N vector.')
        else:
            raise TypeError('Position vector must be a numpy vector with len=3.')

    @property
    def masses(self):
        return self.__masses

    def add(self, pos=np.zeros(3), vel=np.zeros(3), mass=0.0, name=None, radius=0.0, ptype=0,
            a=None, e=0.0, i=0.0, Omega=0.0, omega=0.0, f=0.0, primary=None):
        """
        High-level routine to add a particle to the global set according to the pos/vel or elements.
        :param pos: The 3D position vector of the particle
        :param vel: The 3D velocity vector of the particle
        :param mass: The mass of the particle
        :param name: The name of the particle. Could be either numeric or string
        :param radius: The radius of the particle
        :param ptype: The particle type; 0 = normal particle; 1 = test particle; 2 = low-mass particle
        :param a: Semi-major axis of the particle
        :param e: Eccentricity of the particle
        :param i: Inclination of the particle (radians)
        :param Omega: Longitude of the ascending node of the particle
        :param omega: Argument of periapsis of the particle
        :param f:
        :param primary: The primary object on which the orbital elements are defined
        :return:
        """
        primary_original = primary
        if a is not None:
            # determine the primary body
            primary = self.determine_primary_body(primary)

            # convert orbital elements to Cartesian coordinates
            pos, vel = Tools.from_orbital_elements_to_cartesian(mp=mass, ms=primary.mass,
                                                                semimajor_axis=a,
                                                                eccentricity=e,
                                                                inclination=i,
                                                                longitude_of_ascending_node=Omega,
                                                                argument_of_periapsis=omega,
                                                                true_anomaly=f,
                                                                G=self.CONST_G)
            pos += primary.pos
            vel += primary.vel
        # print name, pos, vel, mass

        self.__positions = np.append(self.__positions, pos)
        self.__velocities = np.append(self.__velocities, vel)
        self.__masses = np.append(self.__masses, np.array([mass]))
        if name is not None:
            self.__names[name] = self.__N
        particle = Particle(mass=mass, pos=self.__positions[3*self.__N:3*self.__N+3],
                            vel=self.__velocities[3*self.__N:3*self.__N+3],
                            name=name, radius=radius, ptype=ptype, primary=primary_original)
        self.__particles.append(particle)
        self.__N += 1

    def add_particle(self, particle):
        """
        Low-level routine to add a particle to the global particle set
        :param particle: The particle to add
        :return:
        """
        if isinstance(particle, Particle) and (particle not in self.particles):
            self.__positions = np.append(self.__positions, particle.pos)
            self.__velocities = np.append(self.__velocities, particle.vel)
            self.__masses = np.append(self.__masses, np.array([particle.mass]))
            if particle.name is not None:
                self.__names[particle.name] = self.__N
            particle.pos = self.__positions[3*self.__N, 3*self.__N+3]
            particle.vel = self.__velocities[3*self.__N+3, 3*self.__N+6]
            self.__particles.append(particle)
            self.__N += 1
        else:
            raise TypeError('Incompatible particle type.')

    def remove_particle(self, particle):
        """
        Low-level routine to remove a particle.
        :param particle: The particle to remove
        :return:
        """
        if isinstance(particle, Particle) and (particle in self.particles):
            pid = self.__particles.index(particle)
            self.__positions = np.delete(self.__positions, range(3 * pid, 3 * pid + 3))
            self.__velocities = np.delete(self.__velocities, range(3 * pid, 3 * pid + 3))
            self.__masses = np.delete(self.__masses, pid)
            self.__particles.remove(particle)
            self.__N -= 1
        else:
            raise TypeError('Incompatible particle type.')

    def merge_particles_inelastically(self, pid1, pid2):
        """
        Merge particles with IDs (pid1, pid2) inelastically, conserving momentum but not energy.
        The less massive particle will be deleted. If some other particles refer to the removed particle as the
        center of mass, then the new center-of-mass will be set to the merged particle.

        :param pid1: The ID of the first particle
        :param pid2: The ID of the second particle
        :return:
        """
        # TODO: ensure that pid1 and pid2 refer to the same particles even if particles are removed
        # TODO: add a hash table for the particle set to quickly find particles uniquely
        try:
            p1 = self.particles[pid1]
            p2 = self.particles[pid2]
        except ValueError:
            return -1
        if p1 is not None and p2 is not None:
            if p1.mass >= p2.mass:
                # merge into p1
                p1.vel = (p1.mass * p1.vel + p2.mass * p2.vel) / (p1.mass + p2.mass)
                self.velocities[3*pid1:3*pid1+3] = p1.vel
                p1.mass = p1.mass + p2.mass
                self.masses[pid1] = p1.mass
                p1.radius = np.power(np.power(p1.radius, 3.0) + np.power(p2.radius, 3.0), 1.0/3)
                self.radii[pid1] = p1.radius
                # search for objects that defines their orbital elements with respect to p2, and change to p1
                for p_id, p in enumerate(self.__particles):
                    if (p.primary is not None) and (self.particles[p.primary] == pid2 or self.particles[p.primary] == p2.name):
                        p.primary = pid1
                self.remove_particle(p2)
                print(("Merging particles inelastically: #%d + #%d ==> #%d" % (pid1, pid2, pid1)))
            else:
                # merge into p2
                p2.vel = (p1.mass * p1.vel + p2.mass * p2.vel) / (p1.mass + p2.mass)
                self.velocities[3*pid2:3*pid2+3] = p2.vel
                p2.mass = p1.mass + p2.mass
                self.masses[pid2] = p2.mass
                p2.radius = np.power(np.power(p1.radius, 3.0) + np.power(p2.radius, 3.0), 1.0/3)
                self.radii[pid2] = p2.radius
                # search for objects that defines their orbital elements with respect to p2, and change to p1
                for p in self.__particles:
                    if (p.primary is not None) and (self.particles[p.primary] == pid1 or self.particles[p.primary] == p1.name):
                        p.primary = pid2
                self.remove_particle(p1)
                print(("Merging particles inelastically: #%d + #%d ==> #%d" % (pid2, pid1, pid2)))
            return 0
        else:
            return -1


    def __getitem__(self, item):
        if isinstance(item, int):
            if item < len(self.__particles):
                return self.__particles[item]
            else:
                raise ValueError('Particle #%d does not exist!' % item)
        elif isinstance(item, string_types):
            if item in self.__names:
                return self.__particles[self.__names[item]]
            else:
                raise ValueError('Particle %s not exist!' % item)
        return None

    def get_center_of_mass(self, subset=None):
        """
        Compute the center-of-mass. If subset is not given, compute the
        center-of-mass of the entire system. Otherwise, compute the COM of
        the subset. Subset is a list of particle IDs in the system.
        """
        com_pos = np.array([0.0, 0.0, 0.0])
        com_vel = np.array([0.0, 0.0, 0.0])
        if subset is None:
            for pid in range(0, self.N):
                com_pos += (self.particles[pid].pos * self.particles[pid].mass)
                com_vel += (self.particles[pid].vel * self.particles[pid].mass)
            com_pos /= np.sum(self.masses)
            com_vel /= np.sum(self.masses)
            return Particle(mass=np.sum(self.masses), pos=com_pos, vel=com_vel)
        else:
            subset_masses = 0.0
            for pid in subset:
                com_pos += (self.particles[pid].pos * self.particles[pid].mass)
                com_vel += (self.particles[pid].vel * self.particles[pid].mass)
                subset_masses += self.particles[pid].mass
            com_pos /= subset_masses
            com_vel /= subset_masses
            return Particle(mass=subset_masses, pos=com_pos, vel=com_vel)

    def calculate_orbital_elements(self, primary=None):
        # calculate the orbital elements
        orbital_elem = np.zeros((self.N, 6))
        if self.N < 2:
            return np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan])

        for pid in range(0, self.N):
            p = self.particles[pid]
            if p.primary is not None:
                # if the particle itself has its own primary object, use its own primary object
                primary = self.determine_primary_body(p.primary)
            else:
                # else use the globally-defined primary
                primary = self.determine_primary_body(primary)
            a, e, f, i, om, Om = Tools.from_cartesian_to_orbital_elements(mp=p.mass, ms=primary.mass,
                                                                          position=np.array([p.x-primary.x, p.y-primary.y, p.z-primary.z]),
                                                                          velocity=np.array([p.vx-primary.vx, p.vy-primary.vy, p.vz-primary.vz]),
                                                                          G=self.CONST_G)
            orbital_elem[pid, :] = np.array([a, e, i, Om, om, f])
        return orbital_elem

    def calculate_aei(self, primary=None):
        # calculate the orbital elements
        orbital_elem = np.zeros((self.N, 3))
        if self.N < 2:
            return np.array([np.nan, np.nan, np.nan])

        for pid in range(0, self.N):
            p = self.particles[pid]
            if p.primary is not None:
                # if the particle itself has its own primary object, use its own primary object
                primary = self.determine_primary_body(p.primary)
            else:
                # else use the globally-defined primary
                primary = self.determine_primary_body(primary)
            a, e, i = Tools.from_cartesian_to_aei(mp=p.mass, ms=primary.mass,
                                                  position=np.array([p.x-primary.x, p.y-primary.y, p.z-primary.z]),
                                                  velocity=np.array([p.vx-primary.vx, p.vy-primary.vy, p.vz-primary.vz]),
                                                  G=self.CONST_G)
            orbital_elem[pid, :] = np.array([a, e, i])
        return orbital_elem

    def determine_primary_body(self, primary):
        if primary is None:
            if self.primary == '#COM#':
                primary = self.get_center_of_mass()
            elif self.primary == '#M_MAX#':
                primary = self.particles[np.argmax(self.masses)]
            elif self.primary == '#M_MIN#':
                primary = self.particles[np.argmin(self.masses)]
            elif isinstance(self.primary, list):
                primary = self.get_center_of_mass(subset=self.primary)
            else:
                primary = self.__getitem__(self.primary)
        elif isinstance(primary, string_types):
            if self.__getitem__(primary) is None:
                raise ValueError('Object with the name %s not found!' % primary)
            else:
                primary = self.__getitem__(primary)
        elif isinstance(primary, list):
            primary = self.get_center_of_mass(subset=primary)
        return primary

    @property
    def energy(self):
        energy = 0.0
        for i in range(0, self.N):
            e_kin = 0.5 * self.masses[i] * np.linalg.norm(self.particles[i].vel) ** 2
            energy += e_kin
            for j in range(0, self.N):
                if i == j:
                    continue
                e_pot = .5 * self.CONST_G * self.masses[i] * self.masses[j] / np.linalg.norm(self.particles[i].pos - self.particles[j].pos)
                energy -= e_pot
        return energy



from setuptools import setup, find_packages
try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension
# from distutils.core import setup, Extension
import sys
import sysconfig
import os 
# name is the package name in pip, should be lower case and not conflict with existing packages
# packages are code source


suffix = sysconfig.get_config_var('EXT_SUFFIX')
if suffix is None:
    suffix = ".so"

extra_link_args=['-fcommon']
if sys.platform == 'darwin':
    from distutils import sysconfig
    vars = sysconfig.get_config_vars()
    vars['LDSHARED'] = vars['LDSHARED'].replace('-bundle', '-shared')
    extra_link_args=['-Wl,-install_name,@rpath/libabie'+suffix]

module_abie = Extension(name = 'libabie',
                        sources = ['abie/integrator_gauss_radau15.c',
                            'abie/integrator_wisdom_holman.c',
                            'abie/integrator_runge_kutta.c',
                            'abie/common.c',
                            'abie/additional_forces.c'],
                        include_dirs = ['abie'],
                        extra_compile_args=['-fstrict-aliasing', '-O3','-std=c99','-march=native','-fPIC', '-shared', '-fcommon'],
                        extra_link_args=extra_link_args,
                        )

setup(name='abie',
      version='0.2',
      description='Alice-Bob Integrator Environment (ABIE), a GPU-accelerated integrator framework for astrophysical N-body simulations',
      url='https://github.com/maxwelltsai/MPA',
      author='Maxwell X. Cai, Javier Roa, Adrian S. Hamers, Nathan W. C. Leigh',
      author_email='maxwellemail@gmail.com',
      license='BSD 2-Clause',
      packages=find_packages(),
      zip_safe=False,
      install_requires=['toml', 'numpy', 'h5py'],
      entry_points={'console_scripts': ['abie = ABIE.abie:main'] },
      ext_modules = [module_abie],
      )



import toml
import numpy as np
import h5py
import os


class DataIO(object):

    def __init__(self, buf_len=1024, output_file_name='data.hdf5', collision_output_file_name='collisions.txt',
                 close_encounter_output_file_name='close_encounters.txt', CONST_G=1):
        self.buf_len = buf_len
        self.buf_initialized = False
        self.buf_t = None  # for the time vector
        self.store_t = 0.0  # the current time of the snapshot data in the buffer
        self.buf_energy = None  # store the total energy
        self.buf_state = None  # for x, y, z, vx, vy, vz
        self.buf_x = None
        self.buf_y = None
        self.buf_z = None
        self.buf_vx = None
        self.buf_vy = None
        self.buf_vz = None
        self.buf_mass = None
        self.buf_radius = None
        self.buf_hashes = None
        self.buf_ptype = None
        self.buf_semi = None
        self.buf_ecc = None
        self.buf_inc = None
        self.buf_cursor = 0
        self.output_file_name = output_file_name
        self.collision_output_file_name = collision_output_file_name
        self.close_encounter_output_file_name = close_encounter_output_file_name
        self.h5_file = None
        self.h5_step_id = 0
        self.CONST_G = CONST_G

    def initialize_buffer(self, n_particles):
        if self.buf_initialized is False:
            buf_len = self.buf_len
            self.buf_t = np.zeros(buf_len) * np.nan
            self.buf_energy = np.zeros(buf_len) * np.nan
            self.buf_mass = np.zeros((buf_len, n_particles)) * np.nan
            self.buf_ptype = np.zeros((buf_len, n_particles), dtype=np.int) * np.nan
            self.buf_hashes = np.zeros((buf_len, n_particles), dtype=np.int) * np.nan
            self.buf_radius = np.zeros((buf_len, n_particles)) * np.nan
            self.buf_x = np.zeros((buf_len, n_particles)) * np.nan
            self.buf_y = np.zeros((buf_len, n_particles)) * np.nan
            self.buf_z = np.zeros((buf_len, n_particles)) * np.nan
            self.buf_vx = np.zeros((buf_len, n_particles)) * np.nan
            self.buf_vy = np.zeros((buf_len, n_particles)) * np.nan
            self.buf_vz = np.zeros((buf_len, n_particles)) * np.nan
            self.buf_semi = np.zeros((buf_len, n_particles)) * np.nan
            self.buf_ecc = np.zeros((buf_len, n_particles)) * np.nan
            self.buf_inc = np.zeros((buf_len, n_particles)) * np.nan
            self.buf_cursor = 0
            # self.h5_step_id = 0
            # remove the previously generated collision / close encounter files
            if os.path.isfile(self.close_encounter_output_file_name):
                os.remove(self.close_encounter_output_file_name)
            if os.path.isfile(self.collision_output_file_name):
                os.remove(self.collision_output_file_name)

            self.buf_initialized = True

    def reset_buffer(self):
        self.buf_initialized = False

    def flush(self):
        """
        Write the simulation data buffer to the HDF5 file unconditionally.
        :return:
        """
        if self.h5_file is None:
            self.h5_file = h5py.File(self.output_file_name, 'w')
            self.h5_file.attrs['G'] = self.CONST_G
        h5_step_group = self.h5_file.create_group('Step#%d' % self.h5_step_id)
        h5_step_group.create_dataset('time', data=self.buf_t[:self.buf_cursor])
        h5_step_group.create_dataset('mass', data=self.buf_mass[:self.buf_cursor])
        h5_step_group.create_dataset('ptype', data=self.buf_ptype[:self.buf_cursor], dtype='int')
        h5_step_group.create_dataset('hash', data=self.buf_hashes[:self.buf_cursor], dtype='int')
        h5_step_group.create_dataset('radius', data=self.buf_radius[:self.buf_cursor])
        h5_step_group.create_dataset('x', data=self.buf_x[:self.buf_cursor])
        h5_step_group.create_dataset('y', data=self.buf_y[:self.buf_cursor])
        h5_step_group.create_dataset('z', data=self.buf_z[:self.buf_cursor])
        h5_step_group.create_dataset('vx', data=self.buf_vx[:self.buf_cursor])
        h5_step_group.create_dataset('vy', data=self.buf_vy[:self.buf_cursor])
        h5_step_group.create_dataset('vz', data=self.buf_vz[:self.buf_cursor])
        h5_step_group.create_dataset('semi', data=self.buf_semi[:self.buf_cursor])
        h5_step_group.create_dataset('ecc', data=self.buf_ecc[:self.buf_cursor])
        h5_step_group.create_dataset('inc', data=self.buf_inc[:self.buf_cursor])
        self.h5_file.flush()
        # reset the cursor
        self.h5_step_id += 1
        self.buf_cursor = 0

    def close(self):
        if self.buf_cursor > 0:
            # there are still some buffer data to be written to the HDF5 file
            self.flush()
        if self.h5_file is not None:
            self.h5_file.close()
            self.h5_file = None

    def store_state(self, t, pos, vel, masses, radii=None, names=None, ptypes=None, a=None, e=None, i=None, energy=None):
        # return if the snapshot time is the same as store_t, which means that this time is already stored
        if self.store_t == t:
            return

        if self.buf_cursor == self.buf_len:
            # the buffer is full, trigger store
            self.flush()
            self.buf_cursor = 0

        self.buf_t[self.buf_cursor] = t
        self.buf_x[self.buf_cursor] = pos[0::3]  # [0, 3, 6, ...]
        self.buf_y[self.buf_cursor] = pos[1::3]  # [1, 4, 7, ...]
        self.buf_z[self.buf_cursor] = pos[2::3]  # [2, 5, 8, ...]
        self.buf_vx[self.buf_cursor] = vel[0::3]  # [0, 3, 6, ...]
        self.buf_vy[self.buf_cursor] = vel[1::3]  # [1, 4, 7, ...]
        self.buf_vz[self.buf_cursor] = vel[2::3]  # [2, 5, 8, ...]
        self.buf_mass[self.buf_cursor] = masses
        if radii is not None:
            self.buf_radius[self.buf_cursor] = radii
        if names is not None:
            self.buf_hashes[self.buf_cursor] = names
        if ptypes is not None:
            self.buf_ptype[self.buf_cursor] = ptypes
        if a is not None:
            self.buf_semi[self.buf_cursor] = a
        if e is not None:
            self.buf_ecc[self.buf_cursor] = e
        if i is not None:
            self.buf_inc[self.buf_cursor] = i
        if energy is not None:
            self.buf_energy[self.buf_cursor] = energy

        self.store_t = t
        self.buf_cursor += 1

    def store(self, buf_t, buf_state, buf_len, n_particles):
        """
        Append `buf_t` and `buf_state` to the instance-wide simulation data buffer. If the buffer is full, it will be
        written to the HDF5 file and then emptied.

        :param buf_t:
        :param buf_state:
        :param buf_len:
        :param n_particles:
        :return:
        """
        if self.buf_len - self.buf_cursor >= buf_len:
            # the buffer storage is sufficient
            self.buf_t[self.buf_cursor:self.buf_cursor+buf_len] = buf_t[:buf_len]
            self.buf_state[self.buf_cursor:self.buf_cursor+buf_len] = buf_state[:buf_len]
            self.buf_cursor += buf_len
        else:
            # full the buffer until it is full, leave the remaining part for the next output
            remaining_buf_len = self.buf_cursor - self.buf_cursor
            self.buf_t[self.buf_cursor:self.buf_len] = buf_t[:remaining_buf_len]
            self.buf_state[self.buf_cursor:self.buf_len] = buf_state[:remaining_buf_len]
            self.buf_cursor = self.buf_len

            # trigger file output if the buffer is full
            self.flush()
            self.buf_cursor = 0

            # store the remaining part into the emptied buffer
            self.buf_t[0:buf_len-remaining_buf_len] = buf_t[remaining_buf_len:buf_len]
            self.buf_state[0:buf_len-remaining_buf_len] = buf_state[remaining_buf_len:buf_len]

        # trigger file output if the buffer is full
        if self.buf_cursor == self.buf_len:
            self.flush()
            self.buf_cursor = 0
            # adjust the buffer length if the default value is too small
            if self.buf_len < buf_len:
                self.buf_len = buf_len

    def store_collisions(self, collision_buffer):
        if self.collision_output_file_name is not None:
            np.savetxt(self.collision_output_file_name, collision_buffer, fmt='%g, %d, %d, %g',
                       header='Time, Particle 1, Particle 2, Distance')

    def store_close_encounters(self, ce_buffer):
        if self.close_encounter_output_file_name is not None:
            np.savetxt(self.close_encounter_output_file_name, ce_buffer, fmt='%g, %d, %d, %g',
                       header='Time, Particle 1, Particle 2, Distance')

    @staticmethod
    def parse_config_file(config_file):
        with open(config_file) as conf_file:
            config = toml.loads(conf_file.read())
            return config  # this is a dict structure

    @staticmethod
    def ic_populate_from_rebound(reb_ic_filename, sim_abie):
        try:
            import rebound
            sim_reb = rebound.Simulation().from_file(reb_ic_filename)
            sim_abie.CONST_G = sim_reb.G
            sim_abie.h = sim_reb.dt
            for particle_id in range(sim_reb.N):
                p = sim_reb.particles[particle_id]
                sim_abie.add(mass=p.m, x=p.x, y=p.y, z=p.z, vx=p.vx, vy=p.vy, vz=p.vz, name=p.hash.value, radius=p.r)
            # data = dict()
            #
            # sim_reb = rebound.Simulation().from_file(reb_ic_filename)
            # data_phy = dict()  # physical parameters
            # data_phy['G'] = sim_reb.G
            # print sim_reb.G
            # data_int = dict()  # integrator
            # data_int['t0'] = 0.0
            # data_int['tf'] = 0.0
            # data_int['h'] = sim_reb.dt
            # data_int['integrator'] = 'GaussRadau15'
            # data_int['acc_method'] = 'ctypes'
            #
            # data_ic = dict()  # initial conditions
            # names = []
            # for particle_id in range(sim_reb.N):
            #     p = sim_reb.particles[particle_id]
            #     pos = [p.x, p.y, p.z]
            #     vel = [p.vx, p.vy, p.vz]
            #     isv = np.append(pos, vel)
            #     isv = np.append(isv, p.m)
            #     p_name = 'p_%05d' % particle_id
            #     data_ic[p_name] = isv.tolist()
            #     names.append(p_name)
            # # sort to preserve the order
            # data['physical_params'] = data_phy
            # data['integration'] = data_int
            # data['initial_conds'] = data_ic
            # data['names'] = names
            #
            # toml.dump(data, open('ss_reb.conf', 'w'))
            # return data
        except ImportError:
            print('Rebound is not installed. Cannot populate IC from rebound.')
            return None

    @staticmethod
    def ic_populate(ic_dict, sim, names=None):
        if names is not None:
            # add the object preserving the name order
            for p_name in names:
                pdata = np.array(ic_dict[p_name])
                sim.add(pos=pdata[0:3], vel=pdata[3:6], mass=pdata[6], name=p_name)
        else:
            # add the objects with random order
            for planet in ic_dict:
                pdata = np.array(ic_dict[planet])
                sim.add(pos=pdata[0:3], vel=pdata[3:6], mass=pdata[6], name=planet)

    @staticmethod
    def toml_generator(state_vec, G, t0, tf, h, masses, filename):
        data = dict()
        data_phy = dict()
        data_phy['G'] = G

        data_int = dict()
        data_int['t0'] = t0
        data_int['tf'] = tf
        data_int['h'] = h

        data_ic = dict()
        ss = ['Sun', 'Mercury', 'Venus', 'Earth', 'Mars', 'Jupiter', 'Saturn', 'Uranus', 'Neptune']
        for i in range(len(ss)):
            pos = state_vec[i*3:i*3+3]
            vel = state_vec[(len(ss)+i)*3:(len(ss)+i)*3+3]
            isv = np.append(pos, vel)
            isv = np.append(isv, masses[i])
            data_ic[ss[i]] = isv.tolist()

        data['physical_params'] = data_phy
        data['integration'] = data_int
        data['initial_conds'] = data_ic

        toml.dump(data, open(filename, 'w'))



from abie.integrator import Integrator
from abie.ode import ODE
import numpy as np
import ctypes
import os

__integrator__ = 'RungeKutta'

# if not os.path.isfile('libode.so'):
#     print('Warning! Shared library libode.so not exsit! Trying to compile.')
#     os.system('cc -fPIC -shared -o libode.so ode.c integrator_gauss_radau15.c -O3')
# lib = ctypes.cdll.LoadLibrary("./libode.so")
# fun = lib.runge_kutta


class RungeKutta(Integrator):

    def __init__(self):
        super(self.__class__, self).__init__()

    def integrate_ctypes(self, to_time=None):
        energy_init = self.calculate_energy()
        dt = min(self.store_dt, self.t_end-self.t)
        self.libabie.initialize_code(self.CONST_G, self.particles.N)
        pos = self.particles.positions.copy()
        vel = self.particles.velocities.copy()
        self.libabie.set_state(pos, vel, self.particles.masses, self.particles.radii, self.particles.N, self.CONST_G)
        energy = self.calculate_energy()
        print(('t = %f, E/E0 = %g' % (self.t, np.abs(energy-energy_init)/energy_init)))
        self.store_state()
        if to_time is not None:
            self.t_end = to_time


        if self.h == 0.0:
            print('ERROR: the timestep for RungeKutta is not set!!! Exiting...')
            import sys
            sys.exit(0)

        while self.t < self.t_end:
            # pos = self.particles.positions.copy()
            # vel = self.particles.velocities.copy()
            # self.libabie.integrator_gauss_radau15(pos, vel, self.particles.masses, self.particles.N, self.CONST_G, self.t, self.t+dt, dt)
            self.libabie.integrator_rk(self.t, self.t+dt, self.h)
            self._t += dt
            self.libabie.get_state(pos, vel, self.particles.masses, self.particles.radii)
            self.particles.positions = pos
            self.particles.velocities = vel
            self.store_state()
            energy = self.calculate_energy()
            print(('t = %f, E/E0 = %g' % (self.t, np.abs(energy-energy_init)/energy_init)))
        self.buf.close()
        return 0

        # dt = min(self.store_dt, self.t_end-self.t)
        # energy_init = self.calculate_energy()
        # self.libabie.initialize_code(self.CONST_G, self.particles.N)
        # if to_time is not None:
        #     self.t_end = to_time
        # while self.t < self.t_end:
        #     pos = self.particles.positions.copy()
        #     vel = self.particles.velocities.copy()
        #     self.libabie.integrator_runge_kutta(pos, vel, self.particles.masses, self.particles.N, self.CONST_G, self.t, self.t+dt, self.h)
        #     self.particles.positions = pos
        #     self.particles.velocities = vel
        #     self._t += dt
        #     self.store_state()
        #     energy = self.calculate_energy()
        #     print('t = %f, E/E0 = %g' % (self.t, np.abs(energy-energy_init)/energy_init))

        # # Allocate dense output
        # npts = int(np.floor((self.t_end - self.t_start) / self.h) + 1)
        #
        # # Initial state
        # x = np.concatenate((self._particles.positions, self._particles.velocities))
        # # Vector of times
        # sol_time = np.linspace(self.t_start, self.t_start + self.h * (npts - 1), npts)
        # energy_init = self.calculate_energy()
        #
        # dxdt = np.zeros(x.size, dtype=np.double)
        # # Launch integration
        # count = 1
        # for t in sol_time[count:]:
        #     fun(ctypes.c_void_p(x.ctypes.data),
        #         ctypes.c_int(x.size / 6),
        #         ctypes.c_double(self.CONST_G),
        #         ctypes.c_double(self.h),
        #         ctypes.c_void_p(self.particles.masses.ctypes.data))
        #     # Store step
        #     self.particles.positions = x[0:self._particles.N * 3]
        #     self.particles.velocities = x[self._particles.N * 3:]
        #     self._t = t
        #     self.store_state()
        #     energy = self.calculate_energy()
        #     print('t = %f, E/E0 = %g' % (self.t, np.abs(energy - energy_init) / energy_init))
        #     count += 1

    def integrate_numpy(self, to_time=None):
        if to_time is not None:
            self.t_end = to_time
        # Allocate dense output
        npts = int(np.floor((self.t_end - self.t_start) / self.h) + 1)

        # Initial state
        x = np.concatenate((self._particles.positions, self._particles.velocities))
        # Vector of times
        sol_time = np.linspace(self.t_start, self.t_start + self.h * (npts - 1), npts)
        energy_init = self.calculate_energy()
        # Launch integration
        count = 1
        for t in sol_time[count:]:
            # Evaluate coefficients
            k1 = ODE.ode_n_body_first_order(x, self.CONST_G, self._particles.masses)
            k2 = ODE.ode_n_body_first_order(x + 0.5 * self.h * k1, self.CONST_G, self._particles.masses)
            k3 = ODE.ode_n_body_first_order(x + 0.5 * self.h * k2, self.CONST_G, self._particles.masses)
            k4 = ODE.ode_n_body_first_order(x + self.h * k3, self.CONST_G, self._particles.masses)

            # Advance the state
            x += (self.h * (k1 + 2 * k2 + 2 * k3 + k4) / 6.0)

            # Store step
            self.particles.positions = x[0:self._particles.N * 3]
            self.particles.velocities = x[self._particles.N * 3:]
            self._t = t
            self.store_state()
            energy = self.calculate_energy()
            print(('t = %f, E/E0 = %g' % (self.t, np.abs(energy - energy_init) / energy_init)))
            count += 1
        self.buf.close()
        return 0
